from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Literal, NamedTuple, Sequence

import torch
import torchmetrics
from torch import Tensor, nn
from torchmetrics import Metric as TorchMetric

from graph_pes.atomic_graph import AtomicGraph, PropertyKey, divide_per_atom
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.misc import uniform_repr
from graph_pes.utils.nn import UniformModuleList

Metric = Callable[[Tensor, Tensor], Tensor]
MetricName = Literal["MAE", "RMSE", "MSE"]


class WeightedLoss:
    def __init__(self, *args, **kwargs):
        # this is now depracated: pass weight directly to the loss object
        raise ImportError(
            "The WeightedLoss class has been removed from graph-pes "
            "as of version 0.0.22. Please now pass loss weights directly "
            "to the loss instances! See the docs for more information: "
            "https://jla-gardner.github.io/graph-pes/fitting/losses.html"
        )


class Loss(nn.Module, ABC):
    """
    A general base class for all loss functions in ``graph-pes``.

    Implementations **must** override:

    * :meth:`forward` to compute the loss value.
    * :meth:`name` to return the name of the loss function.
    * :meth:`required_properties` to return the properties that this loss
      function needs to have available in order to compute its value.

    Additionally, implementations can optionally override:

    * :meth:`pre_fit` to perform any necessary operations before training
      commences.

    Parameters
    ----------
    weight
        a scalar multiplier for weighting the value returned by
        :meth:`forward` as part of a :class:`TotalLoss`.
    is_per_atom
        whether this loss returns a value that is normalised per atom, or not.
        For instance, some metric that acts on ``"forces"`` is naturally
        per-atom, while a metric that acts on ``"energy"``, or the model etc.,
        is not. Specifying this correctly ensures that the effective batch size
        is chosen correctly when averaging over batches.
    """

    def __init__(self, weight, is_per_atom: bool = False):
        super().__init__()
        self.weight = weight
        self.is_per_atom = is_per_atom

    @abstractmethod
    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor | TorchMetric:
        r"""
        Compute the unweighted loss value.

        Note that only :class:`Loss`\ s that return a tensor can be used
        for training: we reserve the use of :class:`TorchMetric`\ s for
        evaluation metrics only.

        :class:`Loss`\ s can act on any of:

        Parameters
        ----------
        model
            The model being trained.
        graph
            The graph (usually a batch) the ``model`` was applied to.
        predictions
            The predictions from the ``model`` for the given ``graph``.
        """

    @property
    @abstractmethod
    def required_properties(self) -> list[PropertyKey]:
        """The properties that are required by this loss function."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of this loss function, for logging purposes."""

    def pre_fit(self, training_data: AtomicGraph):
        """
        Perform any necessary operations before training commences.

        For example, this could be used to pre-compute a standard deviation
        of some property in the training data, which could then be used in
        :meth:`forward`.

        Parameters
        ----------
        training_data
            The training data to pre-fit this loss function to.
        """

    # add type hints to play nicely with mypy
    def __call__(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor | TorchMetric:
        return super().__call__(model, graph, predictions)


class PropertyLoss(Loss):
    r"""
    A :class:`PropertyLoss` instance applies its :class:`Metric` to compare a
    model's predictions to the true values for a given property of a
    :class:`~graph_pes.AtomicGraph`.

    Parameters
    ----------
    property
        The property to apply the loss metric to.
    metric
        The loss metric to use. Defaults to :class:`RMSE`.

    Examples
    --------

    .. code-block:: python

        energy_rmse_loss = PropertyLoss("energy", RMSE())
        energy_rmse_value = energy_rmse_loss(
            predictions,  # a dict of key (energy/force/etc.) to value
            graph.properties,
        )
    """

    def __init__(
        self,
        property: PropertyKey,
        metric: Metric | MetricName | TorchMetric = "RMSE",
        weight: float = 1.0,
    ):
        super().__init__(
            weight,
            is_per_atom=property in ("forces", "local_energies"),
        )
        self.property: PropertyKey = property
        self.metric = parse_metric(metric)

    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor | TorchMetric:
        """
        Computes the loss value.

        Parameters
        ----------
        predictions
            The predictions from the ``model`` for the given ``graph``.
        """

        if isinstance(self.metric, torchmetrics.Metric):
            self.metric.update(
                predictions[self.property],
                graph.properties[self.property],
            )
            return self.metric

        return self.metric(
            predictions[self.property],
            graph.properties[self.property],
        )

    @property
    def name(self) -> str:
        """Get the name of this loss for logging purposes."""
        return f"{self.property}_{_get_metric_name(self.metric)}"

    @property
    def required_properties(self) -> list[PropertyKey]:
        return [self.property]

    def __repr__(self) -> str:
        return uniform_repr(
            self.__class__.__name__,
            self.property,
            metric=self.metric,
        )


class SubLossPair(NamedTuple):
    loss_value: torch.Tensor
    weighted_loss_value: torch.Tensor
    is_per_atom: bool


class TotalLossResult(NamedTuple):
    loss_value: torch.Tensor
    components: dict[str, SubLossPair]


class TotalLoss(torch.nn.Module):
    r"""
    A lightweight wrapper around a collection of losses.

    .. math::

        \mathcal{L}_{\text{total}} = \sum_i w_i \mathcal{L}_i

    where :math:`\mathcal{L}_i` is the :math:`i`-th loss and :math:`w_i` is the
    corresponding weight.

    ``graph-pes`` models are trained by minimising a :class:`TotalLoss` value.

    Parameters
    ----------
    losses
        The collection of losses to aggregate.
    """

    def __init__(self, losses: Sequence[Loss]):
        super().__init__()
        self.losses = UniformModuleList(losses)

    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> TotalLossResult:
        """
        Computes the total loss value.

        Parameters
        ----------
        predictions
            The predictions from the model.
        graph
            The graph (usually a batch) the ``model`` was applied to.
        """

        total_loss = torch.scalar_tensor(0.0, device=graph.Z.device)
        components: dict[str, SubLossPair] = {}

        for loss in self.losses:
            loss_value = loss(model, graph, predictions)
            weighted_loss_value = loss_value * loss.weight

            if not isinstance(loss_value, torch.Tensor):
                raise ValueError(
                    f"Total losses can only be used alongside metrics that "
                    f"return a tensor. Instead, loss {loss.name} returned "
                    f"{type(loss_value)}."
                )

            total_loss += weighted_loss_value
            components[loss.name] = SubLossPair(
                loss_value,
                weighted_loss_value,
                loss.is_per_atom,
            )

        return TotalLossResult(total_loss, components)

    # add type hints to appease mypy
    def __call__(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> TotalLossResult:
        return super().__call__(model, graph, predictions)

    def __repr__(self) -> str:
        return "\n".join(
            ["TotalLoss:"]
            + ["    ".join(str(loss).split("\n")) for loss in self.losses]
        )


############################# CUSTOM LOSSES #############################


class PerAtomEnergyLoss(PropertyLoss):
    r"""
    A loss function that evaluates some metric on the total energy normalised
    by the number of atoms in the structure.

    .. math::
        \mathcal{L} = \text{metric}\left(
        \bigoplus_i \frac{\hat{E}_i}{N_i}, \bigoplus_i\frac{E_i}{N_i} \right)

    where :math:`\hat{E}_i` is the predicted energy for structure :math:`i`,
    :math:`E_i` is the true energy for structure :math:`i`, :math:`N_i`
    is the number of atoms in structure :math:`i` and :math:`\bigoplus_i`
    denotes the cocatenation over all structures in the batch.

    Parameters
    ----------
    metric
        The loss metric to use. Defaults to :class:`RMSE`.
    """

    def __init__(
        self,
        metric: Metric | MetricName | TorchMetric = "RMSE",
        weight: float = 1.0,
    ):
        super().__init__("energy", metric, weight)

    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor | TorchMetric:
        if isinstance(self.metric, torchmetrics.Metric):
            self.metric.update(
                divide_per_atom(predictions["energy"], graph),
                divide_per_atom(graph.properties["energy"], graph),
            )
            return self.metric

        return self.metric(
            divide_per_atom(predictions["energy"], graph),
            divide_per_atom(graph.properties["energy"], graph),
        )

    @property
    def name(self) -> str:
        return f"per_atom_energy_{_get_metric_name(self.metric)}"


class ForceRMSE(PropertyLoss):
    """
    Alias for :class:`PropertyLoss` with ``property="forces"`` and
    ``metric=RMSE``.
    """

    def __init__(self, weight: float = 1.0):
        super().__init__("forces", RMSE(), weight)


## METRICS ##


def parse_metric(metric: Metric | MetricName | TorchMetric | None) -> Metric:
    if isinstance(metric, str):
        return {"MAE": MAE(), "RMSE": RMSE(), "MSE": MSE()}[metric]

    if metric is None:
        return RMSE()

    return metric


class MSE(torch.nn.MSELoss):
    r"""
    Mean squared error metric:

    .. math::
        \frac{1}{N} \sum_i^N \left( \hat{P}_i - P_i \right)^2
    """


class RMSE(torch.nn.MSELoss):
    r"""
    Root mean squared error metric:

    .. math::
        \sqrt{ \frac{1}{N} \sum_i^N \left( \hat{P}_i - P_i \right)^2 }

    .. note::

        Metrics are computed per-batch in the Lightning Trainer.
        When aggregating this metric across multiple batches, we therefore
        get the mean of the per-batch RMSEs, which is not the same as the
        RMSE between all predictions and targets from all batches.

        ``graph-pes`` avoids this issue for all RMSE-based metrics logged as
        ``"{valid|test}/metrics/..._rmse"`` by using a different,
        ``torchmetrics`` based implementation.
    """

    def forward(
        self, input: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        return (super().forward(input, target)).sqrt()

    @property
    def name(self) -> str:
        return "rmse_batchwise"


class MAE(torch.nn.L1Loss):
    r"""
    Mean absolute error metric:

    .. math::
        \frac{1}{N} \sum_i^N \left| \hat{P}_i - P_i \right|
    """


class Huber(torch.nn.HuberLoss):
    r"""
    Huber loss metric:

    .. math::
        L_\delta(y, \hat{y}) = \begin{cases}
            \frac{1}{2}(y - \hat{y})^2 & \text{if } |y - \hat{y}| \leq \delta \\
            \delta (|y - \hat{y}| - \frac{1}{2}\delta) & \text{otherwise}
        \end{cases}
    """

    def __init__(self, delta: float = 0.5):
        super().__init__(delta=delta)

    @property
    def name(self) -> str:
        return f"huber_{self.delta}"


class ScaleFreeHuber(Huber):
    r"""
    A scaled version of the :class:`Huber` loss metric, such that,
    beyond :math:`|y - \hat{y}| = \delta`, the loss is independent of
    :math:`\delta`: this makes tuning the :math:`\delta` value
    easier when several loss components are being used (no need to also
    change the relative weighting of the loss components).

    .. math::
        \mathcal{L}_{\delta}(y, \hat{y}) =
        \frac{1}{\delta} \text{Huber}_\delta(y, \hat{y})
    """

    def forward(
        self, input: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        return super().forward(input, target) / self.delta

    @property
    def name(self) -> str:
        return f"scaled_huber_{self.delta}"


def _get_metric_name(metric: Metric) -> str:
    if hasattr(metric, "name"):
        return metric.name  # type: ignore

    if isinstance(metric, torchmetrics.MeanSquaredError):
        return "mse" if metric.squared else "rmse"

    # if metric is a function, we want the function's name, otherwise
    # we want the metric's class name, all lowercased
    # and without the word "loss" in it
    return (
        getattr(
            metric,
            "__name__",
            metric.__class__.__name__,
        )
        .lower()
        .replace("loss", "")
    )



from __future__ import annotations

from typing import Sequence

import ase
import numpy as np
import torch
from ase.data import chemical_symbols
from sklearn.linear_model import Ridge

from graph_pes.atomic_graph import (
    AtomicGraph,
    number_of_structures,
    sum_per_structure,
    to_batch,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.nn import PerElementParameter

from .logger import logger


def guess_per_element_mean_and_var(
    per_structure_quantity: torch.Tensor,
    batch: AtomicGraph,
    min_variance: float = 0.01,
) -> tuple[dict[int, float], dict[int, float]]:
    r"""
    Guess the per-element mean (:math:`\mu_Z`) and variance (:math:`\sigma_Z^2`)
    of a per-structure quantity using ridge regression under the following assumptions:

    1. the per-structure property, :math:`P`, is a summation over local
       properties of its components atoms: :math:`P = \sum_{i=1}^{N} p_{Z_i}`.
    2. the per-atom properties, :math:`p_{Z_i}`, are independent and identically
       distributed (i.i.d.) for each atom of type :math:`Z_i` according to a
       normal distribution: :math:`p_{Z_i} \sim \mathcal{N}(\mu_{Z_i}, \sigma_{Z_i}^2)`.

    Parameters
    ----------
    per_structure_quantity
        The per-structure quantity to guess the per-element mean and variance of.
    batch
        The batch of graphs to use for guessing the per-element mean and variance.

    Returns
    -------
    means
        A dictionary mapping atomic numbers to per-element means.
    variances
        A dictionary mapping atomic numbers to per-element variances.
    """  # noqa: E501

    # extract the atomic numbers to tensor N such that:
    # N[structure, Z] is the number of atoms of atomic number Z in structure
    unique_Zs = torch.unique(batch.Z)  # (n_Z,)
    N = torch.zeros(number_of_structures(batch), len(unique_Zs))  # (batch, n_Z)
    for i, Z in enumerate(unique_Zs):
        N[:, i] = sum_per_structure((batch.Z == Z).float(), batch)

    if np.linalg.matrix_rank(N.numpy()) < N.shape[1]:
        logger.warning(
            """\
We are attempting to guess the mean per-element
contribution for a per-structure quantity (usually
the total energy). 

However, the composition of the training set is such that 
no unique solution is possible. 

This is probably because you are training on structures
all with the same composition (e.g. all structures are
of the form n H2O). Consider explicitly setting the
per-element contributions if you know them, or
including a variety of structures of different
compositions in the training set.
"""
        )

    # calculate the per-element mean
    # use Ridge rather than LinearRegression to avoid singular matrices
    # when e.g. only one structure contains an atom of a given type...
    ridge = Ridge(fit_intercept=False, alpha=0.00001)
    ridge.fit(N.numpy(), per_structure_quantity)
    mu_Z = torch.tensor(ridge.coef_)
    means = {int(Z): float(mu) for Z, mu in zip(unique_Zs, mu_Z)}

    # calculate the per-element variance
    residuals = per_structure_quantity - N @ mu_Z
    # assuming that the squared residuals are a sum of the independent
    # variances for each atom, we can estimate these variances again
    # using Ridge regression
    ridge.fit(N.numpy(), residuals**2)
    var_Z = ridge.coef_
    # avoid negative variances by clipping to min value
    variances = {
        int(Z): max(float(var), min_variance)
        for Z, var in zip(unique_Zs, var_Z)
    }

    logger.debug(f"Per-element means: {means}")
    logger.debug(f"Per-element variances: {variances}")

    return means, variances


@torch.no_grad()
def add_auto_offset(
    model: GraphPESModel,
    structures: Sequence[AtomicGraph] | Sequence[ase.Atoms],
) -> GraphPESModel:
    from graph_pes.models import AdditionModel, LearnableOffset

    logger.info("""\
Attempting to automatically detect the offset energy for each element.
We do this by first generating predictions for each training structure 
(up to `config.fitting.max_n_pre_fit` if specified). 
This is a slow process! If you already know the reference energies (or the
difference in reference energies if you are fine-tuning an existing model to a
different level of theory), 
we recommend setting `config.fitting.auto_fit_reference_energies` to `False`
and manually specifying a `LearnableOffset` component of your model.

See the "Fine-tuning" tutorial in the docs for more information: 
https://jla-gardner.github.io/graph-pes/quickstart/fine-tuning.html""")

    graphs = [
        AtomicGraph.from_ase(s, model.cutoff.item())
        if isinstance(s, ase.Atoms)
        else s
        for s in structures
    ]

    model.eval()

    predictions = torch.tensor([model.predict_energy(g) for g in graphs])
    references = torch.tensor([g.properties["energy"] for g in graphs])
    diffs = references - predictions
    means, _ = guess_per_element_mean_and_var(diffs, to_batch(graphs))

    if all(abs(mu) < 1e-4 for mu in means.values()):
        logger.info(
            "When attempting to automatically fit the offset energy, "
            "no offset was detected, so skipping."
        )
        return model

    offset_param = PerElementParameter.from_dict(
        **{chemical_symbols[Z]: mu for Z, mu in means.items()},
        requires_grad=True,
    )
    offset_model = LearnableOffset()
    offset_model._offsets = offset_param
    offset_model._has_been_pre_fit.fill_(1)

    if isinstance(model, AdditionModel) and "auto_offset" not in model.models:
        model.models["auto_offset"] = offset_model
    else:
        model = AdditionModel(base=model, auto_offset=offset_model)

    model.train()

    return model



from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from ase.geometry.cell import cell_to_cellpar

from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    edges_per_graph,
    is_batch,
    neighbour_distances,
    neighbour_vectors,
    number_of_atoms,
    structure_sizes,
    to_batch,
    trim_edges,
)
from graph_pes.interfaces.base import InterfaceModel
from graph_pes.utils.misc import voigt_6_to_full_3x3

if TYPE_CHECKING:
    from orb_models.forcefield.conservative_regressor import (
        ConservativeForcefieldRegressor,
    )
    from orb_models.forcefield.direct_regressor import (
        DirectForcefieldRegressor,
    )


def from_graph_pes_to_orb_batch(
    graph: AtomicGraph,
    cutoff: float,
    max_neighbours: int,
):
    from orb_models.forcefield.base import AtomGraphs as OrbGraph

    if not is_batch(graph):
        graph = to_batch([graph])

    graph = trim_edges(graph, cutoff)
    distances = neighbour_distances(graph)

    new_nl, new_offsets = [], []
    for i in range(number_of_atoms(graph)):
        mask = graph.neighbour_list[0] == i
        d = distances[mask]
        if d.numel() == 0:
            continue
        elif d.numel() < max_neighbours:
            new_nl.append(graph.neighbour_list[:, mask])
            new_offsets.append(graph.neighbour_cell_offsets[mask])
        else:
            topk = torch.topk(d, k=max_neighbours, largest=False)
            new_nl.append(graph.neighbour_list[:, mask][:, topk.indices])
            new_offsets.append(graph.neighbour_cell_offsets[mask][topk.indices])

    if new_nl:
        graph = graph._replace(
            neighbour_list=torch.hstack(new_nl),
            neighbour_cell_offsets=torch.vstack(new_offsets),
        )

    node_features = {
        "atomic_numbers": graph.Z.long(),
        "positions": graph.R,
        "atomic_numbers_embedding": torch.nn.functional.one_hot(
            graph.Z, num_classes=118
        ),
        "atom_identity": torch.arange(number_of_atoms(graph)).long(),
    }

    edge_features = {
        "vectors": neighbour_vectors(graph),
        "unit_shifts": graph.neighbour_cell_offsets,
    }

    lattices = []
    for cell in graph.cell.clone().detach():
        lattices.append(
            torch.from_numpy(cell_to_cellpar(cell.cpu().numpy())).float()
        )
    lattice = torch.vstack(lattices).to(graph.R.device)

    graph_features = {
        "cell": graph.cell,
        "pbc": torch.Tensor([False, False, False])
        if not graph.pbc
        else graph.pbc,
        "lattice": lattice,
    }

    return OrbGraph(
        senders=graph.neighbour_list[0],
        receivers=graph.neighbour_list[1],
        n_node=structure_sizes(graph),
        n_edge=edges_per_graph(graph),
        node_features=node_features,
        edge_features=edge_features,
        system_features=graph_features,
        node_targets={},
        edge_targets={},
        system_targets={},
        fix_atoms=None,
        tags=torch.zeros(number_of_atoms(graph)),
        radius=cutoff,
        max_num_neighbors=torch.tensor([max_neighbours]),
        system_id=None,
    ).to(device=graph.R.device, dtype=graph.R.dtype)


class OrbWrapper(InterfaceModel):
    """
    A wrapper around an ``orb-models`` model that converts it into a
    :class:`~graph_pes.GraphPESModel`.

    Parameters
    ----------
    orb
        The ``orb-models`` model to wrap.
    """

    def __init__(self, orb: torch.nn.Module):
        from orb_models.forcefield.conservative_regressor import (
            ConservativeForcefieldRegressor,
        )
        from orb_models.forcefield.direct_regressor import (
            DirectForcefieldRegressor,
        )

        assert isinstance(
            orb, (DirectForcefieldRegressor, ConservativeForcefieldRegressor)
        )
        super().__init__(
            cutoff=orb.system_config.radius,
            implemented_properties=[
                "local_energies",
                "energy",
                "forces",
                "stress",
            ],
        )
        self._orb = orb

    def convert_to_underlying_input(self, graph: AtomicGraph):
        return from_graph_pes_to_orb_batch(
            graph,
            self._orb.system_config.radius,
            self._orb.system_config.max_num_neighbors,
        )

    def raw_forward_pass(
        self,
        input,
        is_batched: bool,
        properties: list[PropertyKey],
    ) -> dict[PropertyKey, torch.Tensor]:
        preds: dict[PropertyKey, torch.Tensor] = self._orb.predict(input)  # type: ignore

        # no easy access to these, but required by the interface, so set to 0
        preds["local_energies"] = torch.zeros_like(
            input.node_features["atomic_numbers"]
        ).float()

        # handle some of the outputs
        if "grad_forces" in preds:
            preds["forces"] = preds.pop("grad_forces")  # type: ignore
        if "grad_stress" in preds:
            preds["stress"] = preds.pop("grad_stress")  # type: ignore
        if "stress" in preds:
            preds["stress"] = voigt_6_to_full_3x3(preds["stress"])

        # underlying orb model returns things in batched format.
        # we want to de-batch things if only a single graph is provided
        if not is_batched:
            preds["energy"] = preds["energy"][0]
            preds["stress"] = preds["stress"].squeeze()

        return preds

    @property
    def orb_model(
        self,
    ) -> "DirectForcefieldRegressor | ConservativeForcefieldRegressor":
        r"""
        Access the underlying ``orb-models`` model.

        One use case of this is to to use ``graph-pes``\ 's fine-tuning
        functionality to adapt an existing ``orb-models`` model to a new
        dataset. You can then re-extract the underlying ``orb-models`` model
        using this property and use it in other ``orb-models`` workflows.
        """
        return self._orb


def orb_model(name: str = "orb-v3-direct-20-omat") -> OrbWrapper:
    """
    Load a pre-trained Orb model, and convert it into a
    :class:`~graph_pes.GraphPESModel`.

    See the `orb-models <https://github.com/orbital-materials/orb-models>`_
    repository for more information on the available models.
    As of 2025-04-11, the following are available:

    * ``"orb-v3-conservative-20-omat"``
    * ``"orb-v3-conservative-inf-omat"``
    * ``"orb-v3-direct-20-omat"``
    * ``"orb-v3-direct-inf-omat"``
    * ``"orb-v3-conservative-20-mpa"``
    * ``"orb-v3-conservative-inf-mpa"``
    * ``"orb-v3-direct-20-mpa"``
    * ``"orb-v3-direct-inf-mpa"``
    * ``"orb-v2"``
    * ``"orb-d3-v2"``
    * ``"orb-d3-sm-v2"``
    * ``"orb-d3-xs-v2"``
    * ``"orb-mptraj-only-v2"``

    Parameters
    ----------
    name: str
        The name of the model to load.
    """
    import torch._functorch.config
    from orb_models.forcefield import pretrained

    torch._functorch.config.donated_buffer = False

    orb = pretrained.ORB_PRETRAINED_MODELS[name](device="cpu")
    for param in orb.parameters():
        param.requires_grad = True
    return OrbWrapper(orb)



import ase
import numpy as np
import pytest
import torch
from ase.build import bulk, molecule
from orb_models.forcefield import atomic_system
from orb_models.forcefield.base import batch_graphs

from graph_pes.atomic_graph import AtomicGraph, to_batch
from graph_pes.interfaces._orb import orb_model
from graph_pes.utils.misc import full_3x3_to_voigt_6


@pytest.fixture(params=["orb-v3-direct-20-omat", "orb-v3-conservative-20-omat"])
def wrapped_orb(request):
    return orb_model(request.param)


def test_single_isolated_structure(wrapped_orb):
    atoms = molecule("H2O")

    g = AtomicGraph.from_ase(atoms, cutoff=wrapped_orb.cutoff.item())
    our_preds = wrapped_orb.forward(g)

    assert our_preds["energy"].shape == tuple()
    assert our_preds["forces"].shape == (3, 3)

    orb_g = atomic_system.ase_atoms_to_atom_graphs(
        atoms, wrapped_orb.orb_model.system_config
    )
    orb_preds = wrapped_orb.orb_model.predict(orb_g)

    torch.testing.assert_close(our_preds["energy"], orb_preds["energy"][0])
    torch.testing.assert_close(
        our_preds["forces"],
        orb_preds["forces"]
        if "forces" in orb_preds
        else orb_preds["grad_forces"],
    )


def test_single_periodic_structure(wrapped_orb):
    atoms = bulk("Cu")
    g = AtomicGraph.from_ase(atoms, cutoff=wrapped_orb.cutoff.item())
    our_preds = wrapped_orb.forward(g)

    assert our_preds["energy"].shape == tuple()
    assert our_preds["forces"].shape == (1, 3)
    assert our_preds["stress"].shape == (3, 3)

    orb_g = atomic_system.ase_atoms_to_atom_graphs(
        atoms, wrapped_orb.orb_model.system_config
    )
    orb_preds = wrapped_orb.orb_model.predict(orb_g)

    torch.testing.assert_close(
        our_preds["energy"],
        orb_preds["energy"][0],
        atol=1e-4,
        rtol=1e-4,
    )
    torch.testing.assert_close(
        our_preds["forces"],
        orb_preds["forces"]
        if "forces" in orb_preds
        else orb_preds["grad_forces"],
        atol=1e-4,
        rtol=1e-4,
    )
    torch.testing.assert_close(
        full_3x3_to_voigt_6(our_preds["stress"]),
        orb_preds["stress"][0]
        if "stress" in orb_preds
        else orb_preds["grad_stress"][0],
        atol=1e-3,
        rtol=1e-3,
    )


def test_batched(wrapped_orb):
    atoms = bulk("Cu").repeat(2)
    rng = np.random.RandomState(42)
    atoms.positions += rng.uniform(-0.1, 0.1, atoms.positions.shape)

    N = len(atoms)
    B = 2

    g = AtomicGraph.from_ase(atoms, cutoff=wrapped_orb.cutoff.item())
    batch = to_batch([g] * B)
    our_preds = wrapped_orb.forward(batch)

    assert our_preds["energy"].shape == (B,)
    assert our_preds["forces"].shape == (B * N, 3)
    assert our_preds["stress"].shape == (B, 3, 3)

    orb_g = atomic_system.ase_atoms_to_atom_graphs(
        atoms, wrapped_orb.orb_model.system_config
    )
    orb_g = batch_graphs([orb_g] * B)
    orb_preds = wrapped_orb.orb_model.predict(orb_g)

    torch.testing.assert_close(our_preds["energy"], orb_preds["energy"])
    torch.testing.assert_close(
        our_preds["forces"],
        orb_preds["forces"]
        if "forces" in orb_preds
        else orb_preds["grad_forces"],
    )
    torch.testing.assert_close(
        full_3x3_to_voigt_6(our_preds["stress"]),
        orb_preds["stress"]
        if "stress" in orb_preds
        else orb_preds["grad_stress"],
    )


def test_single_atom(wrapped_orb):
    wrapped_orb.ase_calculator().calculate(ase.Atoms("H"))



from __future__ import annotations

import warnings
from contextlib import redirect_stdout
from itertools import chain
from pathlib import Path
from typing import Literal

import requests
import torch

from graph_pes import AtomicGraph
from graph_pes.atomic_graph import PropertyKey, is_batch
from graph_pes.interfaces.base import InterfaceModel
from graph_pes.utils.misc import MAX_Z


class ZToOneHot(torch.nn.Module):
    def __init__(self, elements: list[int]):
        super().__init__()
        self.z_to_index: torch.Tensor
        self.register_buffer("z_to_index", torch.full((MAX_Z + 1,), -1))
        for i, z in enumerate(elements):
            self.z_to_index[z] = i
        self.num_classes = len(elements)

    def forward(self, Z: torch.Tensor) -> torch.Tensor:
        if (Z > 118).any():
            not_allowed: list[int] = torch.unique(Z[Z > 118]).tolist()
            raise ValueError(
                "ZToOneHot received an atomic number is too large: "
                f"{not_allowed}."
            )

        indices = self.z_to_index[Z]

        if (indices < 0).any():
            not_allowed: list[int] = torch.unique(Z[indices < 0]).tolist()
            raise ValueError(
                "ZToOneHot received an atomic number that is not in the model's"
                f" element list: {not_allowed}. Please ensure the model was "
                "trained with all elements present in the input graph."
            )

        return torch.nn.functional.one_hot(indices, self.num_classes)


def _atomic_graph_to_mace_input(
    graph: AtomicGraph,
    z_to_one_hot: ZToOneHot,
) -> dict[str, torch.Tensor]:
    batch = graph.batch
    if batch is None:
        batch = torch.zeros_like(graph.Z)

    ptr = graph.ptr
    if ptr is None:
        ptr = torch.tensor([0, graph.Z.shape[0]])

    cell = graph.cell.unsqueeze(0) if not is_batch(graph) else graph.cell

    _cell_per_edge = cell[batch[graph.neighbour_list[0]]]  # (E, 3, 3)
    _shifts = torch.einsum(
        "kl,klm->km", graph.neighbour_cell_offsets, _cell_per_edge
    )  # (E, 3)
    data = {
        "node_attrs": z_to_one_hot.forward(graph.Z).to(graph.R.dtype),
        "positions": graph.R.clone().detach(),
        "cell": cell.clone().detach(),
        "edge_index": graph.neighbour_list,
        "unit_shifts": graph.neighbour_cell_offsets,
        "shifts": _shifts,
        "batch": batch,
        "ptr": ptr,
    }
    return {k: v.to(graph.Z.device) for k, v in data.items()}


class MACEWrapper(InterfaceModel):
    """
    Converts any MACE model from the `mace-torch <https://github.com/ACEsuit/mace-torch>`__
    package into a :class:`~graph_pes.GraphPESModel`.

    You can use this to drive MD using LAMMPS, fine-tune MACE models,
    or any functionality that ``graph-pes`` provides.

    Parameters
    ----------
    model
        The MACE model to wrap.

    Examples
    --------
    >>> mace_torch_model = ...  # create your MACE model any-which way
    >>> from graph_pes.interfaces._mace import MACEWrapper
    >>> graph_pes_model = MACEWrapper(mace_torch_model)  # convert to graph-pes
    >>> graph_pes_model.predict_energy(graph)
    torch.Tensor([123.456])
    >>> from graph_pes.utils.calculator import GraphPESCalculator
    >>> calculator = GraphPESCalculator(graph_pes_model)
    >>> calculator.calculate(ase_atoms)

    """

    def __init__(self, model: torch.nn.Module):
        super().__init__(
            model.r_max.item(),  # type: ignore
            implemented_properties=[
                "local_energies",
                "energy",
                "forces",
                "stress",
                "virial",
            ],
        )
        self.model = model
        self.z_to_one_hot = ZToOneHot(self.model.atomic_numbers.tolist())  # type: ignore

    def convert_to_underlying_input(
        self, graph: AtomicGraph
    ) -> dict[str, torch.Tensor]:
        return _atomic_graph_to_mace_input(graph, self.z_to_one_hot)

    def raw_forward_pass(
        self,
        input: dict[str, torch.Tensor],
        is_batched: bool,
        properties: list[PropertyKey],
    ) -> dict[PropertyKey, torch.Tensor]:
        MACE_KEY_MAPPING: dict[str, PropertyKey] = {
            "node_energy": "local_energies",
            "energy": "energy",
            "forces": "forces",
            "stress": "stress",
            "virials": "virial",
        }

        raw_predictions = self.model.forward(
            input,
            training=self.training,
            compute_force="forces" in properties,
            compute_stress="stress" in properties,
            compute_virials="virial" in properties,
        )

        predictions: dict[PropertyKey, torch.Tensor] = {}
        for key, value in raw_predictions.items():
            if key in MACE_KEY_MAPPING:
                property_key = MACE_KEY_MAPPING[key]
                if property_key in properties and value is not None:
                    predictions[property_key] = value

        if not is_batched:
            for p in ["energy", "stress", "virial"]:
                if p in properties:
                    predictions[p] = predictions[p].squeeze()

        return predictions

    def predict(
        self,
        graph: AtomicGraph,
        properties: list[PropertyKey],
    ) -> dict[PropertyKey, torch.Tensor]:
        # override this predict property to use the underlying
        # mace-torch mechanisms for calculating forces etc.
        input = self.convert_to_underlying_input(graph)
        is_batched = is_batch(graph)
        return self.raw_forward_pass(input, is_batched, properties)


def _fix_dtype(model: torch.nn.Module, dtype: torch.dtype) -> None:
    for tensor in chain(
        model.parameters(),
        model.buffers(),
    ):
        if tensor.dtype.is_floating_point:
            tensor.data = tensor.data.to(dtype)


def _get_dtype(
    precision: Literal["float32", "float64"] | None,
) -> torch.dtype:
    if precision is None:
        return torch.get_default_dtype()
    return {"float32": torch.float32, "float64": torch.float64}[precision]


def mace_mp(
    model: str = "small",
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    """
    Donwload a MACE-MP model and convert it for use with ``graph-pes``.

    Internally, we use the `foundation_models <https://mace-docs.readthedocs.io/en/latest/guide/foundation_models.html>`__
    functionality from the `mace-torch <https://github.com/ACEsuit/mace-torch>`__ package.

    Please cite the following if you use this model:

    - MACE-MP by Ilyes Batatia, Philipp Benner, Yuan Chiang, Alin M. Elena,
      Dávid P. Kovács, Janosh Riebesell, et al., 2023, arXiv:2401.00096
    - MACE-Universal by Yuan Chiang, 2023, Hugging Face, Revision e5ebd9b,
      DOI: 10.57967/hf/1202, URL: https://huggingface.co/cyrusyc/mace-universal
    - Matbench Discovery by Janosh Riebesell, Rhys EA Goodall, Philipp Benner,
      Yuan Chiang, Alpha A Lee, Anubhav Jain, Kristin A Persson, 2023,
      arXiv:2308.14920

    As of 10th April 2025, the following models are available: ``["small", "medium", "large", "medium-mpa-0", "small-0b", "medium-0b", "small-0b2", "medium-0b2", "medium-0b3", "large-0b2", "medium-omat-0"]``

    Parameters
    ----------
    model
        The size of the MACE-MP model to download.
    precision
        The precision of the model. If ``None``, the default precision
        of torch will be used (you can set this when using ``graph-pes-train``
        via ``general/torch/dtype``)
    """  # noqa: E501
    from mace.calculators.foundations_models import mace_mp as mace_mp_impl

    dtype = _get_dtype(precision)
    precision_str = {torch.float32: "float32", torch.float64: "float64"}[dtype]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=FutureWarning)
        with redirect_stdout(None):
            mace_torch_model = mace_mp_impl(
                model,
                device="cpu",
                default_dtype=precision_str,
                return_raw_model=True,
            )

    assert isinstance(mace_torch_model, torch.nn.Module)
    _fix_dtype(mace_torch_model, dtype)
    return MACEWrapper(mace_torch_model)


def mace_off(
    model: Literal["small", "medium", "large"],
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    """
    Download a MACE-OFF model and convert it for use with ``graph-pes``.

    If you use this model, please cite the relevant paper by Kovacs et.al., arXiv:2312.15211

    Parameters
    ----------
    model
        The size of the MACE-OFF model to download.
    precision
        The precision of the model.
    """  # noqa: E501
    from mace.calculators.foundations_models import mace_off

    dtype = _get_dtype(precision)
    precision_str = {torch.float32: "float32", torch.float64: "float64"}[dtype]

    mace_torch_model = mace_off(
        model,
        device="cpu",
        default_dtype=precision_str,
        return_raw_model=True,
    )
    assert isinstance(mace_torch_model, torch.nn.Module)
    _fix_dtype(mace_torch_model, dtype)
    # un freeze all parameters
    for p in mace_torch_model.parameters():
        p.requires_grad = True
    return MACEWrapper(mace_torch_model)


def go_mace_23(
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    """
    Download the `GO-MACE-23 model <https://doi.org/10.1002/anie.202410088>`__
    and convert it for use with ``graph-pes``.

    .. note::

        This model is only for use on structures containing Carbon, Hydrogen and
        Oxygen. Attempting to use on structures with other elements will raise
        an error.

    If you use this model, please cite the following:

    .. code-block:: bibtex

        @article{El-Machachi-24,
            title = {Accelerated {{First-Principles Exploration}} of {{Structure}} and {{Reactivity}} in {{Graphene Oxide}}},
            author = {{El-Machachi}, Zakariya and Frantzov, Damyan and Nijamudheen, A. and Zarrouk, Tigany and Caro, Miguel A. and Deringer, Volker L.},
            year = {2024},
            journal = {Angewandte Chemie International Edition},
            volume = {63},
            number = {52},
            pages = {e202410088},
            doi = {10.1002/anie.202410088},
        }

    """  # noqa: E501

    return _mace_from_url(
        "https://github.com/zakmachachi/GO-MACE-23/raw/refs/heads/main/models/fitting/potential/iter-12-final-model/go-mace-23.pt",
        "GO-MACE-23",
        precision,
    )


def egret(
    model: Literal["egret-1", "egret-1t", "egret-1e"] = "egret-1",
) -> MACEWrapper:
    """
    Download an `Egret <https://arxiv.org/abs/2504.20955>`__ model
    and convert it for use with ``graph-pes``.

    Use the ``egret-1`` model via the Python API:

    .. code-block:: python

        from graph_pes.interfaces._mace import egret
        model = egret("egret-1")

    or :doc:`fine-tune <../quickstart/fine-tuning>` it on your own data using
    the :doc:`graph-pes-train <../cli/graph-pes-train/root>` command:

    .. code-block:: yaml

        model:
            +egret: {model: "egret-1e"}

        data:
            ...
        # etc.

    If you use this model, please cite the following:

    .. code-block:: bibtex

        @misc{Mann-25-04,
            title = {
                Egret-1: {{Pretrained Neural Network
                Potentials For Efficient}} and
                {{Accurate Bioorganic Simulation}}
            },
            author = {
                Mann, Elias L. and Wagen, Corin C.
                and Vandezande, Jonathon E. and Wagen, Arien M.
                and Schneider, Spencer C.
            },
            year = {2025},
            number = {arXiv:2504.20955},
            doi = {10.48550/arXiv.2504.20955},
        }

    As of 1st May 2025, the following models are available: ``["egret-1", "egret-1t", "egret-1e"]``
    Parameters
    ----------
    model
        The model to download.
    """  # noqa: E501

    urls = {
        "egret-1": "https://github.com/rowansci/egret-public/raw/b1b4c1261315b38f0dd3f6ec0fce891a9119ffe0/compiled_models/EGRET_1.model",
        "egret-1t": "https://github.com/rowansci/egret-public/raw/b1b4c1261315b38f0dd3f6ec0fce891a9119ffe0/compiled_models/EGRET_1T.model",
        "egret-1e": "https://github.com/rowansci/egret-public/raw/b1b4c1261315b38f0dd3f6ec0fce891a9119ffe0/compiled_models/EGRET_1E.model",
    }

    return _mace_from_url(urls[model], model)


def _mace_from_url(
    url: str,
    model_name: str,
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    dtype = _get_dtype(precision)
    file_name = f"{model_name.replace(' ', '-')}.pt"

    save_path = Path.home() / ".graph-pes" / file_name
    save_path.parent.mkdir(parents=True, exist_ok=True)

    if not save_path.exists():
        print(f"Downloading {model_name} model to {save_path}")
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for bad status codes

        with open(save_path, "wb") as file:
            file.write(response.content)

    print(f"Loading {model_name} model from {save_path}")
    mace_torch_model = torch.load(
        save_path, weights_only=False, map_location=torch.device("cpu")
    )
    for p in mace_torch_model.parameters():
        p.data = p.data.to(dtype)
    model = MACEWrapper(mace_torch_model)

    return model
